import { createFilter } from 'rollup-pluginutils';
import { findConfigFile, sys, getParsedCommandLineOfConfigFile, createProgram, nodeModuleNameResolver, createSourceFile, ScriptTarget, ModifierFlags, SyntaxKind, isFunctionDeclaration, isInterfaceDeclaration, isExportDeclaration, isExportAssignment, isImportDeclaration, isPropertySignature, isNamedImports, isTypeReferenceNode, isIdentifier, isLiteralExpression, getCombinedModifierFlags } from 'typescript';
import path from 'path';

const MARKER = "0";
let IDs = 1;
/**
 * Create a new `Program` for the given `node`:
 */
function createProgram$1(node) {
    return withStartEnd({
        type: "Program",
        sourceType: "module",
        body: [],
    }, node);
}
/**
 * Create an export for `id`:
 * `export default id`
 */
function createDefaultExport(node, range) {
    const id = createIdentifier(node, range);
    return withStartEnd({
        type: "ExportDefaultDeclaration",
        declaration: id,
    }, range);
}
/**
 * Create an export for `id`:
 * `export { id }`
 */
function createExport(node, range) {
    const id = createIdentifier(node, range);
    return withStartEnd({
        type: "ExportNamedDeclaration",
        declaration: null,
        specifiers: [
            {
                type: "ExportSpecifier",
                exported: id,
                local: id,
            },
        ],
    }, range);
}
/**
 * Creates a reference to `id`:
 * `_ = ${id}`
 */
function createReference(id) {
    return {
        type: "AssignmentPattern",
        left: {
            type: "Identifier",
            name: String(IDs++),
        },
        right: id,
    };
}
function createIdentifier(node, range = node) {
    return withStartEnd({
        type: "Identifier",
        name: node.text || String(node.escapedText),
    }, range);
}
/**
 * Create a new Declaration and Scope for `id`:
 * `function ${id}(_ = MARKER) {}`
 */
function createDeclaration(id, range) {
    const start = getStart(range);
    return withStartEnd({
        type: "FunctionDeclaration",
        id: withStartEnd({
            type: "Identifier",
            name: id.text,
        }, { start, end: start }),
        params: [createReference(createIdentifier(id))],
        body: { type: "BlockStatement", body: [] },
    }, range);
}
/**
 * Mark the nested `range` to be removed, by creating dead code:
 * `_ = () => {MARKER}`
 */
function removeNested(range) {
    return createReference({
        type: "FunctionExpression",
        id: null,
        body: withStartEnd({
            type: "BlockStatement",
            body: [
                withStartEnd({
                    type: "ExpressionStatement",
                    expression: { type: "Identifier", name: MARKER },
                }, range),
            ],
        }, {
            start: getStart(range) - 1,
            end: range.end + 1,
        }),
        params: [],
    });
}
function getStart({ start, pos }) {
    return typeof start === "number" ? start : pos || 0;
}
function withStartEnd(node, range = { start: 0, end: 0 }) {
    return Object.assign(node, {
        start: getStart(range),
        end: range.end,
    });
}
function matchesModifier(node, flags) {
    const nodeFlags = getCombinedModifierFlags(node);
    return (nodeFlags & flags) === flags;
}

class Transformer {
    constructor(sourceFile) {
        this.ast = createProgram$1(sourceFile);
        for (const stmt of sourceFile.statements) {
            this.convertStatement(stmt);
        }
    }
    transform() {
        return this.ast;
    }
    pushStatement(node) {
        if (node) {
            this.ast.body.push(node);
        }
    }
    maybeMarkAsExported(node, id) {
        if (matchesModifier(node, ModifierFlags.ExportDefault)) {
            const start = node.pos;
            this.pushStatement(createDefaultExport(id, { start, end: start }));
        }
        else if (matchesModifier(node, ModifierFlags.Export)) {
            const start = node.pos;
            this.pushStatement(createExport(id, { start, end: start }));
        }
    }
    removeExportModifier(node) {
        const ret = [];
        for (const mod of node.modifiers || []) {
            if (mod.kind === SyntaxKind.ExportKeyword) {
                const start = node.getStart();
                const end = mod.end + 1;
                ret.push(removeNested({ start, end }));
            }
        }
        return ret;
    }
    convertStatement(node) {
        if (isFunctionDeclaration(node)) {
            return this.convertFunctionDeclaration(node);
        }
        if (isInterfaceDeclaration(node)) {
            return this.convertInterfaceDeclaration(node);
        }
        if (isExportDeclaration(node) || isExportAssignment(node)) {
            return this.convertExportDeclaration(node);
        }
        if (isImportDeclaration(node)) {
            return this.convertImportDeclaration(node);
        }
        console.log({ code: node.getFullText() });
        throw new Error(`unsupported node type`);
    }
    createDeclaration(id, range) {
        const decl = createDeclaration(id, range);
        this.pushStatement(decl);
        return decl.params;
    }
    convertFunctionDeclaration(node) {
        if (!node.name) {
            console.log(node);
            throw new Error(`FunctionDeclaration should have a name`);
        }
        this.maybeMarkAsExported(node, node.name);
        const body = this.createDeclaration(node.name, node);
        body.push(...this.removeExportModifier(node));
        for (const param of node.parameters) {
            if (param.type) {
                this.convertTypeNode(param.type, body);
            }
        }
        if (node.type) {
            this.convertTypeNode(node.type, body);
        }
    }
    convertInterfaceDeclaration(node) {
        if (!node.name) {
            console.log(node);
            throw new Error(`InterfaceDeclaration should have a name`);
        }
        this.maybeMarkAsExported(node, node.name);
        const body = this.createDeclaration(node.name, node);
        body.push(...this.removeExportModifier(node));
        for (const heritage of node.heritageClauses || []) {
            for (const type of heritage.types) {
                body.push(createReference(this.convertExpression(type.expression)));
            }
        }
        for (const member of node.members) {
            if (isPropertySignature(member) && member.type) {
                this.convertTypeNode(member.type, body);
            }
        }
    }
    convertExportDeclaration(node) {
        if (isExportAssignment(node)) {
            this.pushStatement(withStartEnd({
                type: "ExportDefaultDeclaration",
                declaration: this.convertExpression(node.expression),
            }, node));
            return;
        }
        const source = node.moduleSpecifier ? this.convertExpression(node.moduleSpecifier) : undefined;
        if (!node.exportClause) {
            this.pushStatement(withStartEnd({
                type: "ExportAllDeclaration",
                source,
            }, node));
        }
        else {
            this.pushStatement(withStartEnd({
                type: "ExportNamedDeclaration",
                declaration: null,
                specifiers: node.exportClause ? node.exportClause.elements.map(e => this.convertExportSpecifier(e)) : [],
                source,
            }, node));
        }
    }
    convertImportDeclaration(node) {
        if (!node.importClause || !node.importClause.namedBindings) {
            throw new Error(`ImportDeclaration should have imports`);
        }
        this.pushStatement(withStartEnd({
            type: "ImportDeclaration",
            specifiers: this.convertNamedImportBindings(node.importClause.namedBindings),
            source: this.convertExpression(node.moduleSpecifier),
        }, node));
    }
    convertNamedImportBindings(node) {
        if (isNamedImports(node)) {
            return node.elements.map(el => {
                const local = createIdentifier(el.name);
                const imported = el.propertyName ? createIdentifier(el.propertyName) : local;
                return {
                    type: "ImportSpecifier",
                    local,
                    imported,
                };
            });
        }
        return [
            {
                type: "ImportNamespaceSpecifier",
                local: createIdentifier(node.name),
            },
        ];
    }
    convertExportSpecifier(node) {
        const exported = createIdentifier(node.name);
        return {
            type: "ExportSpecifier",
            exported: exported,
            local: node.propertyName ? createIdentifier(node.propertyName) : exported,
        };
    }
    convertTypeNode(node, body) {
        if (isTypeReferenceNode(node)) {
            body.push(createReference(this.convertEntityName(node.typeName)));
        }
    }
    convertEntityName(node) {
        if (isIdentifier(node)) {
            return createIdentifier(node);
        }
        return withStartEnd({
            type: "MemberExpression",
            computed: false,
            object: this.convertEntityName(node.left),
            property: createIdentifier(node.right),
        }, 
        // TODO: clean up all the `start` handling!
        { start: node.getStart(), end: node.end });
    }
    convertExpression(node) {
        if (isLiteralExpression(node)) {
            return { type: "Literal", value: node.text };
        }
        if (isIdentifier(node)) {
            return createIdentifier(node);
        }
        console.log({ code: node.getFullText() });
        throw new Error(`Unknown Expression`);
    }
}

let SOURCEMAPPING_URL = "sourceMa";
SOURCEMAPPING_URL += "ppingURL";
const SOURCEMAPPING_URL_RE = new RegExp(`^//#\\s+${SOURCEMAPPING_URL}=.+\\n?`, "m");
const COMPILERCACHE = new Map();
function createCompiler(options) {
    const file = path.basename(options.tsconfig);
    const configFileName = findConfigFile(options.tsconfig, sys.fileExists, path.extname(options.tsconfig) ? file : undefined);
    if (!configFileName) {
        throw new Error(`rollup-plugin-dts: Couldn't find tsconfig file`);
    }
    const compilerOptions = Object.assign({}, options.compilerOptions, { noEmitOnError: false, noEmit: false, skipLibCheck: true, declaration: true, 
        // hm, there are still issues with this, I couldn’t get it to work locally
        // See https://github.com/Microsoft/TypeScript/issues/25662
        declarationMap: false });
    let diagnostic;
    const configParseResult = getParsedCommandLineOfConfigFile(configFileName, compilerOptions, Object.assign({}, sys, { onUnRecoverableConfigFileDiagnostic(d) {
            diagnostic = d;
        } }));
    if (!configParseResult) {
        console.log(diagnostic);
        throw new Error(`rollup-plugin-dts: Couldn't process compiler options`);
    }
    const { fileNames, options: parsedCompilerOptions } = configParseResult;
    return {
        parsedCompilerOptions,
        compiler: createProgram(fileNames, parsedCompilerOptions),
    };
}
function getCachedCompiler(options) {
    const cacheKey = JSON.stringify(options);
    function lazyCreate() {
        let compiler = COMPILERCACHE.get(cacheKey);
        if (!compiler) {
            compiler = createCompiler(options);
            COMPILERCACHE.set(cacheKey, compiler);
        }
        return compiler;
    }
    return {
        get compiler() {
            return lazyCreate().compiler;
        },
        resolve(importee, importer) {
            const { parsedCompilerOptions } = lazyCreate();
            const result = nodeModuleNameResolver(importee, importer, parsedCompilerOptions, sys);
            if (result.resolvedModule && result.resolvedModule.resolvedFileName) {
                return result.resolvedModule.resolvedFileName;
            }
            return null;
        },
        transform(fileName) {
            const { compiler } = lazyCreate();
            const sourceFile = compiler.getSourceFile(fileName);
            let dtsFilename = "";
            let code = "";
            let map = `{"mappings": ""}`;
            const baseFileName = fileName.slice(0, -path.extname(fileName).length);
            compiler.emit(sourceFile, (fileName, data) => {
                if (fileName === `${baseFileName}.d.ts`) {
                    dtsFilename = fileName;
                    code = data.replace(SOURCEMAPPING_URL_RE, "").trim();
                }
                if (fileName === `${baseFileName}.d.ts.map`) {
                    map = data;
                }
            });
            const dtsSource = createSourceFile(dtsFilename, code, ScriptTarget.Latest, true);
            const converter = new Transformer(dtsSource);
            const ast = converter.transform();
            // NOTE(swatinem):
            // hm, typescript generates `export default` without a declare,
            // but rollup moves the `export default` to a different place, which leaves
            // the function declaration without a `declare`.
            // Well luckily both words have the same length, haha :-D
            code = code.replace(/(export\s+)default(\s+function)/m, "$1declare$2");
            return { code, ast, map };
        },
    };
}

// @ts-ignore
function dts(options = {}) {
    const filter = createFilter(options.include || ["*.ts+(|x)", "**/*.ts+(|x)"], options.exclude || []);
    const compiler = getCachedCompiler({
        tsconfig: options.tsconfig || process.cwd(),
        compilerOptions: options.compilerOptions || {},
    });
    return {
        name: "dts",
        resolveId(importee, importer) {
            if (!importer) {
                return null;
            }
            importer = importer.split("\\").join("/");
            return compiler.resolve(importee, importer);
        },
        async transform(_code, id) {
            if (!filter(id)) {
                return;
            }
            return compiler.transform(id);
        },
    };
}

export default dts;
